/**
 * engine.js
 */

'use strict';

export default class Engine {

  /**
   * 构造函数
   * @param activities 流程步骤
   * @param data 表单数据
   */
  constructor(activities, data) {
    this.activities = activities.map(item => new Activity(item, this));
    this.data = data;
  }

  getNext(id) {

    const activity = this.findActivity(id);

    // 优先执行子步骤
    const child = activity.child();
    if (child) {
      return child;
    }

    // 执行兄弟步骤
    const next = activity.next();
    if (next) {
      return next;
    }

    // 没有设置下一步节点，那么回到上一层的节点继续
    return this.getParent(activity);
  }

  getParent(activity) {

    let prev = activity.prev();

    // 如果前序步骤是是兄弟步骤，那就继续找上一层，直到找到父步骤
    if (prev.isBrother(activity)) {
      return this.getParent(prev);
    }

    // 否则prev就是父步骤，回到父步骤继续
    return prev.next();
  }

  // 计算所有条件，找到满足条件的分支
  calculate(condition) {
    return condition.find(or => {
      return or.reduce((prev, and) => prev && Engine.compare(and), true);
    });
  }

  findActivity(id) {
    return this.activities.find(item => item.id === id);
  }

  static compare(item) {

    const a = this.data[item.key], b = item.value;

    switch (item.operator) {
      case 'eq':
        return a === b;
      case 'gt':
        return a > b;
      case 'gte':
        return a >= b;
      case 'lt':
        return a < b;
      case 'lte':
        return a <= b;
      case 'include':
        return a.includes(b);
      case 'exclude':
        return !a.includes(b);
      case 'regex':
        return a.match(new RegExp(b));
    }

    return false;
  }

}

export class Activity {
  constructor(activity, engine) {
    this.id = activity.id;
    this.transitions = activity.transitions;
    this.engine = engine;
  }

  // 下一个步骤（用condition为空来判断）
  next() {
    const transition = this.transitions.find(item => !item.condition);
    return this.engine.findActivity(transition.to);
  }

  // 上一个步骤
  prev() {
    return this.engine.activities.find(item => {
      return item.transitions.find(transition => transition.to === this.id);
    });
  }

  // 判断当前的步骤是给定步骤的前序步骤
  isBrother(to) {
    return this.next().id === to.id;
  }

  // 获取所有分支条件
  conditions() {
    return this.transitions.filter(item => item.condition);
  }

  // 获取分支步骤
  child() {

    // 遍历所有条件，确认条件是否满足
    const transition = this.conditions().find(item => this.engine.calculate(item.condition));
    if (transition) {
      return this.engine.findActivity(transition.to);
    }

    return null;
  }
}

